"""
*****************************************************************
Licensed Materials - Property of IBM
(C) Copyright IBM Corp. 2019. All Rights Reserved.
US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*****************************************************************
"""

import tensorflow as tf
import ddl
import numpy as np
import sys
from os import environ
import argparse

def all_reduce_test (data, num_grads, rank,
                     input_dtype,
                     test=True,
                     average=True,
                     test_name="",
                     delta=1e-5,
                     expected_grad=[]):
    if ddl.rank() == 0:
        print ("Running test for " + test_name)

    #Define the model
    x        = tf.placeholder(dtype=input_dtype, shape=[None, 1], name="X")
    y_act    = tf.placeholder(dtype=input_dtype, shape=[None, 1], name="Y")
    y_pred = x
    for i in range(num_grads):
        y_pred += tf.Variable(initial_value=data * i, name="w"+str(i))

    cost = tf.reduce_mean(tf.pow(y_pred - y_act, 2))

    optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.0001)
    grads_and_vars = optimizer.compute_gradients(cost, tf.trainable_variables())

    # Initialize the variables (i.e. assign their default value)
    init = tf.global_variables_initializer()

    #Prepare dummy data to pass to the model
    x_data = data
    rand_data = np.linspace(0, rank, num=data.shape[0])
    y_act_data = rand_data.reshape(data.shape) + data

    # Start session
    g_and_v = ()
    with tf.Session() as sess:
        #Run the initializer
        sess.run(init)

        g_and_v = sess.run(grads_and_vars, feed_dict={x:x_data, y_act:y_act_data})

        grads, vars = zip(*g_and_v)
        grads = np.ravel(grads)

        if test:
            g_and_v = sess.run(ddl._grads_reduce(grads_and_vars, average=average),\
                                        feed_dict={x:x_data, y_act:y_act_data})
            grads, vars = zip(*g_and_v)
            grads = np.ravel(grads)
            if not np.allclose(grads, expected_grad, rtol=delta):
                if ddl.rank() == 0:
                    print (str(ddl_rank) + ": printing (grad_reduce avg="+ str(average) + ") grads")
                    print (grads)
                    print ("ERROR: gradient from allreduce is not same as expected")
                    print (expected_grad)
                sys.exit(1)

    tf.reset_default_graph()
    return grads

def calculate_grad (data_len, num_grads, dtype, average=True):
    result_grads = np.zeros(num_grads * data_len, dtype=dtype)

    for i in range(ddl.size()):
        x_data = np.ones((data_len,1), dtype=dtype) * (i + 1)
        grads = all_reduce_test (x_data, num_grads, rank=i+1, input_dtype=dtype, test=False)

        result_grads += grads

    if average:
        result_grads /= ddl.size()

    return result_grads

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='DDL allreduce op testcase')
    parser.add_argument('--ddl_options', type=str,
                        default='-mode b:2x1 -dump_iter 100',
                        help='DDL options. e.g. "-mode b:2x1 -dump_iter 100"')
    parser.add_argument('--grad_size', type=int,
                        default=1024,
                        help='Size of each gradient')
    parser.add_argument('--num_grads', type=int,
                        default=4,
                        help='Number of gradients')
    parser.add_argument('--iterations', type=int,
                        default=10,
                        help='Number of iterations to perform')
    FLAGS, unparsed = parser.parse_known_args()

    if environ.get('DDL_OPTIONS') is None:
        #Need to explicitly call ddl.init
        ddl.init(FLAGS.ddl_options)

    ddl_rank = ddl.rank() + 1
    data_len = FLAGS.grad_size
    num_grads = FLAGS.num_grads

    #Figure out the expected gradients for each of the test cases
    expected_grad_sum_16 = calculate_grad(data_len=data_len, num_grads=num_grads, dtype=np.float16, average=False)
    expected_grad_avg_16 = calculate_grad(data_len=data_len, num_grads=num_grads, dtype=np.float16, average=True)

    expected_grad_sum_32 = calculate_grad(data_len=data_len, num_grads=num_grads, dtype=np.float32, average=False)
    expected_grad_avg_32 = calculate_grad(data_len=data_len, num_grads=num_grads, dtype=np.float32, average=True)

    if ddl.rank() == 0:
        print ("expected_grad_sum_16 = " + str(expected_grad_sum_16))
        print ("expected_grad_avg_16 = " + str(expected_grad_avg_16))
        print ("expected_grad_sum_32 = " + str(expected_grad_sum_32))
        print ("expected_grad_avg_32 = " + str(expected_grad_avg_32))

    float16_arr = np.ones((data_len,1), dtype=np.float16) * ddl_rank
    float32_arr = np.ones((data_len,1), dtype=np.float32) * ddl_rank

    for i in range(FLAGS.iterations):
        #TestCase1: Call allreduce with average for float16 dtype
        all_reduce_test(data=float16_arr, num_grads=num_grads, rank=ddl_rank, input_dtype=np.float16, test=True, average=True, delta=1e-3, test_name="float16 with average=True", expected_grad=expected_grad_avg_16)

        #TestCase2: Call allreduce with sum for float16 dtype
        all_reduce_test(data=float16_arr, num_grads=num_grads, rank=ddl_rank, input_dtype=np.float16, test=True, average=False, delta=1e-3, test_name="float16 with average=False", expected_grad=expected_grad_sum_16)

        #TestCase3: Call allreduce with average for float32 dtype
        all_reduce_test(data=float32_arr, num_grads=num_grads, rank=ddl_rank, input_dtype=np.float32, test=True, average=True, test_name="float32 with average=True", expected_grad=expected_grad_avg_32)

        #TestCase4: Call allreduce with sum for float32 dtype
        all_reduce_test(data=float32_arr, num_grads=num_grads, rank=ddl_rank, input_dtype=np.float32, test=True, average=False, test_name="float32 with average=False", expected_grad=expected_grad_sum_32)

